import torch
import math
from magbox.Wait_bar import Wait_bar
import warnings

class eq_solver:
    def __init__(self,odeFcn, t_span, y0: torch.Tensor, solver_name, options):
        self.device=y0.device
        self.dtype=y0.dtype
        self.t2pi=torch.tensor(2*math.pi,dtype=self.dtype,device=self.device)
        self._ode_options(options)
        self._ode_initial(odeFcn, t_span, y0)
        self._tableau(solver_name)
    def _ode_options(self, options):
        device=self.device
        dtype=self.dtype
        if options is None:
            options = {}
        
        # Initialize options
        self.waitbar = options.get('waitbar', True)
        
        # Extract odeset options
        rtol = options.get('rel_tol', torch.tensor(1e-3,device=device,dtype=dtype))
        atol = options.get('abs_tol', torch.tensor(1e-6,device=device,dtype=dtype))
        max_failures = options.get('max_consecutive_failures', torch.tensor(10,device=device,dtype=torch.int64))
        refine = options.get('refine', 4)
        max_step = options.get('max_step', torch.tensor(2**32-1,device=device, dtype=torch.int64))

        self.rtol=torch.as_tensor(rtol, device=device, dtype=dtype)
        self.atol=torch.as_tensor(atol, device=device, dtype=dtype)
        self.max_failures=torch.as_tensor(max_failures, device=device, dtype=torch.int64)
        self.refine=refine
        self.max_step = torch.as_tensor(max_step, device=device, dtype=torch.int64)

    def _ode_initial(self, odeFcn, t_span, y0):
        device=self.device
        dtype=self.dtype
        self.t0 = t_span[0]
        self.t_final = t_span[-1]
        self.t_dir = torch.sign(self.t_final - self.t0)
        # step size constraints
        h_min = 16 * torch.finfo(dtype).eps
        h_min=torch.tensor(h_min, dtype=dtype,device=device)
        safe_h_max = 16.0 * torch.finfo(dtype).eps * torch.max(torch.abs(self.t0), torch.abs(self.t_final))
        default_h_max = torch.max(0.1 * torch.abs(self.t_final - self.t0), safe_h_max)
        h_max = torch.min(torch.abs(self.t_final - self.t0), torch.min(self.max_step, default_h_max))

        t = self.t0.clone()
        y = y0.clone()


        n_t_span=t_span.shape[0]
        n_eq=y0.shape
        S=torch.tensor(0,device=device,dtype=dtype)
        chunk=0
        refine=self.refine
        if n_t_span > 2:
            output_pos = 1  # output at t_span points
        elif self.refine <= 1:
            output_pos = 2  # computed points
        else:
            output_pos = 3  # computed points, with refinement
            S = torch.linspace(1/refine, 1 - 1/refine, refine - 1, dtype=dtype, device=device)
        # Initialize output arrays
        if n_t_span > 2:
            t_out = torch.zeros(n_t_span, dtype=dtype, device=device)
            y_out = torch.zeros((n_t_span,)+n_eq, dtype=dtype, device=device)
        else:
            chunk = min(max(100, 50 * refine), refine + (2**13) // n_eq)
            t_out = torch.zeros(chunk, dtype=dtype, device=device)
            y_out = torch.zeros((int(chunk),)+n_eq, dtype=dtype, device=device)
        self.S=S
        self.chunk=chunk
        self.h_min=h_min
        self.h_max=h_max
        self.t=t
        self.y=y
        self.n_t_span=n_t_span
        self.n_eq=n_eq
        self.t_span=t_span
        self.output_pos=output_pos
        self.t_out=t_out
        self.y_out=y_out
        self.ode_fcn=odeFcn

        # return t0, t_final, t_dir, y0, h_min, h_max, t, y0, output_pos, t_out, y_out, chunk, S
    def _tableau(self, solver_name):
        device=self.device
        dtype=self.dtype
        if solver_name == 'RK45':
            self.alpha=torch.tensor([1 / 5, 3 / 10, 4 / 5, 8 / 9, 1., 1.], 
                                    dtype=dtype,device=device)
            self.beta=[
                torch.tensor([1 / 5], dtype=dtype,device=device).view(-1,1),
                torch.tensor([3 / 40, 9 / 40], dtype=dtype,device=device).view(-1,1),
                torch.tensor([44 / 45, -56 / 15, 32 / 9], dtype=dtype,device=device).view(-1,1),
                torch.tensor([19372 / 6561, -25360 / 2187, 64448 / 6561, -212 / 729], dtype=dtype,device=device).view(-1,1),
                torch.tensor([9017 / 3168, -355 / 33, 46732 / 5247, 49 / 176, -5103 / 18656], dtype=dtype,device=device).view(-1,1),
                torch.tensor([35 / 384, 0, 500 / 1113, 125 / 192, -2187 / 6784, 11 / 84], dtype=dtype,device=device).view(-1,1)
            ]
            self.c_error=torch.tensor([71/57600, 0.0, -71/16695, 71/1920, -17253/339200, 22/525, -1/40]
                # [71 / 86400, 0, -142 / 50085, 71 / 2880, -5751 / 169600, 44 / 1575, -1 / 60],
                                      , dtype=dtype,device=device).view(-1,1)
            self.order=5
            self.interp_coeff=torch.tensor([
                [1, -183/64, 37/12, -145/128],
                [0,0,0,0],
                [0,1500/371, -1000/159, 1000/371],
                [0, -125/32, 125/12, -375/64],
                [0, 9477/3392, -729/106, 25515/6784],
                [0,-11/7, 11/3, -55/28],
                [0, 3/2, -4, 5/2]
            ], dtype=dtype,device=device)
        else:
            raise ValueError(f'Unknown solver: {solver_name}')
    def _get_parameters(self):
        return self.t_out, self.y_out, self.t, self.y, self.dtype, self.device, self.h_max, self.h_min, self.rtol, self.atol, self.max_failures, self.t_final, self.t0, self.t_dir, self.t_span, self.waitbar, self.alpha, self.beta, self.c_error, self.ode_fcn, self.S, self.chunk, self.output_pos, self.n_t_span, self.n_eq, self.refine, self.order, self.interp_coeff
    @staticmethod
    def _one_step(f_list: torch.Tensor, y_list: torch.Tensor, t, y, h, ode_fcn, alpha,beta):
        t_new = t + h
        # f_list shape: [b, N, ...]  y_list shape: [7, N, ...] 
        # f_list[0] has been filled with k1
        for i, (alpha_i, beta_i) in enumerate(zip(alpha,beta)):
            # Use data up to the current stage i+1 (indices 0 to i) to compute calculating stage
            # beta_i has size i+1
            # Efficiently compute dy using pre-allocated f_list slice
            dy = torch.einsum('s...,sj->...', f_list[:i+1], beta_i)
            
            y_next = y + h * dy
            y_list[i+1] = y_next

            if alpha_i == 1. :
                ti=t_new
            else:
                ti=t+alpha_i*h

            k_next = ode_fcn(ti, y_next)
            f_list[i+1] = k_next

        return t_new, y_list, f_list
    @staticmethod
    def _get_err(f_list: torch.Tensor, c_error: torch.Tensor, h, y, y_new, atol):
        err = torch.einsum('s...,si->...', f_list, c_error)
        err /= torch.max(torch.max(y.abs(),y_new.abs()), atol)
        err = h *  torch.max(err.abs())
        return err.item()
    
    @staticmethod
    def _get_outputs(output_pos, interp_fun, t, S, refine, t_new, y_new, y, h, f_list, interp_coeff, dtype, device, n_t_span, t_span, next_idx, t_dir):
        if output_pos ==2: # computed points
                nout_new = 1
                t_out_new = t_new.unsqueeze(0)
                y_out_new = y_new
        elif output_pos ==3: # computed points, with refinement
            t_ref = t + (t_new - t) * S
            nout_new = refine
            t_out_new = torch.cat([t_ref, t_new.unsqueeze(0)])
            y_interp = interp_fun(t_ref, t, y, h, f_list, interp_coeff)
            y_out_new = torch.cat([y_interp, y_new.unsqueeze(0)], dim=0) 
        else:
            nout_new = 0
            t_out_new = torch.tensor([], dtype=dtype, device=device)
            y_out_new = torch.tensor([], dtype=dtype, device=device)
            
            while next_idx < n_t_span:
                if t_dir * (t_new - t_span[next_idx]) < 0:
                    break
                nout_new += 1
                t_out_new = torch.cat([t_out_new, t_span[next_idx].unsqueeze(0)])
                if t_span[next_idx] == t_new:
                    y_out_new = torch.cat([y_out_new, y_new.unsqueeze(0)], dim=0)
                else:
                    y_interp = interp_fun(t_span[next_idx], t, y, h, f_list, interp_coeff)
                    y_out_new = torch.cat([y_out_new, y_interp], dim=0)
                next_idx += 1
        return nout_new, t_out_new, y_out_new, next_idx
    @staticmethod
    def _interp_fun(t_interp: torch.Tensor, t: torch.Tensor, y: torch.Tensor,h: torch.Tensor ,f_list: torch.Tensor, interp_coeff, is_dy: bool=True)-> torch.Tensor:
        """
        Interpolation function for Dormand-Prince method.
        """
        max_order = interp_coeff.shape[1] # interp_coeff.shape = [f_order, t_order]
        s = (t_interp - t) / h
        s = s.reshape(1,-1) #[1,t_l]
        s_list = torch.zeros((max_order, s.shape[1]), dtype=f_list.dtype, device=f_list.device) #[t_order,t_l]
        s_list[0] = s 
        if max_order > 1:
            for jj in range(max_order-1):
                s_list[jj+1] = s**(jj+2) # [t_order,t_l]
        s_coeff = interp_coeff @ s_list #[f_order, t_order] @ [t_order,t_l] = [f_order, t_l]
        dy = torch.einsum('s...,st->t...', f_list, s_coeff)  # [f_order,...] @ [f_order, t_l] = [t_l,...]
        if is_dy:
            y_interp = y.unsqueeze(0) + h * dy
        else:
            y_interp = y.unsqueeze(0) + dy  # [1,...] + [t_l,...] = [t_l,...]
        return y_interp
    
    @staticmethod
    def _optimal_step_size(last_step, error_ratio, order, failedflag, safety=0.8, ifactor=4.0, dfactor=0.2):
        """Calculate the optimal size for the next step."""
        if failedflag:
            return last_step
        else: 
            ifactor=torch.as_tensor(ifactor,dtype=last_step.dtype,device=last_step.device)
            dfactor=torch.as_tensor(dfactor,dtype=last_step.dtype,device=last_step.device)
            if error_ratio == 0:
                return last_step * ifactor
            if error_ratio < 1:
                dfactor = torch.ones((), dtype=last_step.dtype, device=last_step.device)
            error_ratio = error_ratio.type_as(last_step)
            exponent = torch.as_tensor(order, dtype=last_step.dtype, device=last_step.device).reciprocal()
            factor = torch.min(ifactor, torch.max(safety / error_ratio ** exponent, dfactor))
            return last_step * factor
    @staticmethod
    def _step_after_nofailed(h_min,h_abs,rtol,err,order):
        return torch.max(h_min, h_abs * max(0.1, 0.8 * (rtol / err) ** (1/order)))
    
    def _after_process(self,y: torch.Tensor) -> torch.Tensor:
        return y/self.t2pi
    
    def run(self,bar: Wait_bar):
        finished = False

        next_idx = 1 # for t_span output
        n_failures = 0
        integration_failed= False
        err_history=[]
        n_steps = 0

        n_out=0

        t_out, y_out, t, y, dtype, device, h_max, h_min, rtol, atol, max_failures, t_final, t0, t_dir, t_span, waitbar, alpha, beta, c_error, ode_fcn, S, chunk, output_pos, n_t_span, n_eq, refine, order, interp_coeff = self._get_parameters()

        n_beta=len(beta)
        f1=ode_fcn(t,y)
        
        n_calls=1

        t_out[n_out]=t
        y_out[n_out,...]=y

        h = torch.min(h_max, torch.max(h_min, 0.1 * torch.abs(t_final -t0)))
        h_abs=torch.abs(h)

        # Pre-allocate buffers for RK45 stages (7 stages)
        stage_shape = (n_beta+1,) + y.shape
        f_list_buffer = torch.zeros(stage_shape, dtype=dtype, device=device)
        y_list_buffer = torch.zeros(stage_shape, dtype=dtype, device=device)

        while not finished:
            h_abs=torch.min(h_max, torch.max(h_min, h_abs))
            h = t_dir*h_abs
            if h_abs > torch.abs(t_final-t):
                h = t_final - t
                h_abs = torch.abs(h)
                finished = True
            failed = False
            y_list_buffer[0] = y
            f_list_buffer[0] = f1
            while True:
                t_new, y_list_buffer, f_list_buffer = self._one_step(f_list_buffer, y_list_buffer, t, y, h, ode_fcn, alpha, beta)
                n_calls += n_beta
                y_new = y_list_buffer[n_beta]
                err = self._get_err(f_list_buffer, c_error, h_abs, y, y_new, atol)
                # step acceptance
                accept_step = err <= rtol
                if accept_step:
                    n_failures = 0
                if h_abs <= h_min:
                    accept_step = True
                    n_failures +=1
                    failed=True
                    if n_failures >= max_failures:
                        bar.close(waitbar)
                        warnings.warn(
                            f"Step size reached minimum hmin = {h_min.item():.2e} at t={t.item():.2e}, but still cannot satisfy tolerance. "
                            f"Current error: {err:.2e}, Required tolerance: {rtol:.2e}. "
                            f"This may indicate a stiff ODE or overly strict tolerances. "
                            f"Consider using a stiff solver or relaxing tolerances.",
                            RuntimeWarning,
                            stacklevel=2
                        )
                        finished = True
                        integration_failed = True
                        break
                else:
                    n_failures = 0 # Reset if we're still above hmin
                if accept_step:
                    err_history.append(err)
                    break
                else:
                    if failed:
                        h_abs = torch.max(h_min, 0.5* h_abs)
                    else:
                        failed = True 
                        h_abs =self._step_after_nofailed(h_min,h_abs,rtol,err,order)
                    h = t_dir * h_abs
                    y_list_buffer[0] = y
                    f_list_buffer[0] = f1
                    finished = False
            n_steps += 1
            if integration_failed:
                break
            # Update waitbar if enabled
            bar.update(t_new, h, waitbar, finished)
            # output
            nout_new, t_out_new, y_out_new, next_idx = self._get_outputs(output_pos,self._interp_fun, t, S, refine, t_new, y_new, y, h, f_list_buffer, interp_coeff, dtype, device, n_t_span, t_span, next_idx,t_dir)
            y_out_new = self._after_process(y_out_new)
            # Store output
            if nout_new > 0:
                old_nout = n_out
                n_out += nout_new

                if n_out+1 > t_out.shape[0]:
                    extra = max(chunk, nout_new)
                    tout_new_temp = torch.zeros(t_out.shape[0] + extra, dtype=dtype, device=device)
                    tout_new_temp[:t_out.shape[0]] = t_out
                    t_out = tout_new_temp
                    
                    yout_new_temp = torch.zeros((y_out.shape[0] + extra,)+n_eq, dtype=dtype, device=device)
                    yout_new_temp[:y_out.shape[0],...] = y_out
                    y_out = yout_new_temp
            
                t_out[old_nout+1:n_out+1] = t_out_new
                y_out[old_nout+1:n_out+1,...] = y_out_new
            h_abs = self._optimal_step_size(h_abs, err/rtol, order, failed)
        
            t=t_new
            y=y_new
            y=self._after_process(y)
            f1 = f_list_buffer[order+1]

        bar.close(waitbar)

        t_out= t_out[:n_out+1]
        y_out = y_out[:n_out+1,...]
        stats = {'n_calls': n_calls,
            'n_steps': n_steps,
            'n_output': n_out+1,
            'integration': not integration_failed}
        err_info = {
            'err_history': err_history,
            'max_step_error': max(err_history) if err_history else 0.0
        }
        return t_out, y_out, stats, err_info



from dataclasses import dataclass, field
from typing import Union, Callable, Optional
import torch

@dataclass
class Lattice:
    size: list = field(default_factory=list)
    type: str = field(default_factory=str)
    periodic: Union[bool, list[bool]] = field(default_factory=lambda: True)
    J_direction: Union[int, None] = field(default_factory=lambda: None)
    force_periodic: bool = field(default_factory=lambda: False)

@dataclass
class Vars:
    B: float = field(default_factory=lambda: 0.0)
    B_dir: list[float] = field(default_factory=lambda: [0.0, 0.0, 1.0])
    J: Union[float, list[float]] = field(default_factory=lambda: 1.0)
    K1_dir: list[float] = field(default_factory=lambda: [0.0, 0.0, 1.0])
    K1: float = field(default_factory=lambda: 1.0)
    D: float = field(default_factory=lambda: 0.0)
    custom_heff: Optional[Callable[[object, object, object], torch.Tensor]] = field(default_factory=lambda: None)
    custom_kernel: tuple =field(default_factory=lambda: (1,1))
    require_vars_grad: bool = field(default_factory=lambda: False)



import torch
from typing import Callable, Tuple, Dict, Any, Optional
import warnings
import math
from .Wait_bar import Wait_bar
from .initial import Lattice
from .solver.sde_solver import sde_solver
from .solver.sde3_solver import sde3_solver
from .solver.sde3_fix_solver import sde3_fix_solver
from .solver.eq_solver import eq_solver
from .solver.eq3_solver import eq3_solver
from .solver.eq3_fix_solver import eq3_fix_solver

def get_data_type(type):
    if type=="f32":
        data_type=torch.float32
    elif type=="f64":
        data_type=torch.float64
    elif type=='f16':
        data_type=torch.float16
    else:
        raise ValueError("type must be f16, f32 or f64")
    return data_type
    
def get_device(device):
    if device=="gpu":
        if torch.cuda.is_available():
            return torch.device("cuda")
        else:
            print("CUDA is not available, using CPU instead.")
            return torch.device("cpu")
    elif device=="cpu":
        return torch.device("cpu")
    else:
        raise ValueError("device must be 'cpu' or 'gpu'")
def get_Jmtx(lattice_type: Lattice,device=torch.device("cuda"),data_type=torch.float32) -> torch.Tensor:
    l_type=lattice_type.type

    if l_type=="square":
        N=lattice_type.size
        N_dim=len(N)
        if N_dim==1:
            N=N+[1,1]
        elif N_dim==2:
            N=N+[1]
            
        pd=lattice_type.periodic #type: ignore
        Force_keep_pd = lattice_type.force_periodic
        if is_bool_or_single_bool_list(pd):
            pd=create_bool_list(pd, N) 
            if N_dim==1:
                pd[1]=False
                pd[2]=False
            elif N_dim==2:
                pd[2]=False
        pd: list[bool] = pd
        direction=lattice_type.J_direction
        totalN=math.prod(N)
        N1=N[0]
        N2=N[1]
        N3=N[2]
        pd_warning_flag = [False, False, False]
        forced_dim=[]
        for i in range(3):
            if N[i] <=2 and pd[i]:
                if not Force_keep_pd:
                    pd[i] = False
                forced_dim.append(i)
                pd_warning_flag[i] = True
        if Force_keep_pd and any(pd_warning_flag):
            warnings.warn(f"Periodic boundary condition in the dimension with length equal to 1 or 2 is set to True. This may cause errors in the simulation. Forced dimension:{forced_dim}", UserWarning, stacklevel=2)
            
        if direction is None:
            # 所有方向的耦合
            v = torch.ones(3 * totalN - N1 * N2 - N2 * N3 - N3 * N1, dtype=data_type,device=device) / 2
            
            # get backward coupling, direction 1
            i = torch.arange(1, totalN)  
            back_boundary = (i % N1 == 0)
            i = i[~back_boundary]
            j = i.clone()  
            
            # get right coupling, direction 2
            itmp = torch.arange(1, totalN)
            right_boundary = ((itmp - 1) % (N1 * N2) - N1 * (N2 - 1)) >= 0
            itmp = itmp[~right_boundary]
            i = torch.cat([i, itmp])
            j = torch.cat([j, itmp + N1 - 1])  # 调整索引
            
            # get bottom coupling, direction 3
            i = torch.cat([i, torch.arange(1, totalN - N1 * N2 + 1)])
            j = torch.cat([j, torch.arange(N1 * N2 , totalN )])
            
            if pd[0]:  # periodic boundary condition in direction 1
                back_forward_i = torch.arange(1, totalN + 1, N1)
                back_forward_j = torch.arange(N1-1, totalN, N1)
                i=torch.cat([i, back_forward_i])
                j=torch.cat([j, back_forward_j])
                v= torch.cat([v, torch.ones(N2 * N3, dtype=data_type,device=device) / 2])
            if pd[1]:  # periodic boundary condition in direction 2
                left_right_i = torch.arange(1, totalN + 1)
                tmpbd = ((left_right_i - 1) % (N1 * N2)) >= N1
                left_right_i = left_right_i[~tmpbd]
                left_right_j = torch.arange(0, totalN)
                tmpbd = (left_right_j  % (N1 * N2)) < N1 * (N2 - 1)
                left_right_j = left_right_j[~tmpbd]
                i=torch.cat([i, left_right_i])
                j=torch.cat([j, left_right_j])
                v= torch.cat([v, torch.ones(N1 * N3, dtype=data_type,device=device) / 2])
            if pd[2]:  # periodic boundary condition in direction 3
                up_down_i = torch.arange(1, N1 * N2 + 1)
                up_down_j = torch.arange(N1 * N2 * (N3 - 1), totalN)
                i=torch.cat([i, up_down_i])
                j=torch.cat([j, up_down_j])
                v= torch.cat([v, torch.ones(N1 * N2, dtype=data_type,device=device) / 2])

        else:
            # 只有一个方向的耦合
            if direction==0: # backward耦合（x方向）
                i = torch.arange(1, totalN)
                back_boundary = (i % N1 == 0)
                i = i[~back_boundary]
                j = i.clone()  # 调整索引
                v = torch.ones(len(i), dtype=data_type,device=device) / 2
                if pd[0]:  # periodic boundary condition in direction 1
                    back_forward_i = torch.arange(1, totalN + 1, N1)
                    back_forward_j = torch.arange(N1-1, totalN , N1)
                    i=torch.cat([i, back_forward_i])
                    j=torch.cat([j, back_forward_j])

                v = torch.ones(len(i), dtype=data_type,device=device) / 2
            elif direction==1: # right耦合（y方向）
                i = torch.arange(1, totalN)
                right_boundary = ((i - 1) % (N1 * N2) - N1 * (N2 - 1)) >= 0
                i = i[~right_boundary]
                j = i + N1 - 1  # 调整索引
                if pd[1]:  # periodic boundary condition in direction 2
                    left_right_i = torch.arange(1, totalN + 1)
                    tmpbd = ((left_right_i - 1) % (N1 * N2)) >= N1
                    left_right_i = left_right_i[~tmpbd]
                    left_right_j = torch.arange(0, totalN )
                    tmpbd = (left_right_j % (N1 * N2)) < N1 * (N2 - 1)
                    left_right_j = left_right_j[~tmpbd]
                    i=torch.cat([i, left_right_i])
                    j=torch.cat([j, left_right_j])
                v = torch.ones(len(i), dtype=data_type,device=device) / 2
            elif direction == 2:  # bottom耦合（z方向）
                i = torch.arange(1, totalN - N1 * N2 + 1)
                j = torch.arange(N1 * N2, totalN)
                if pd:  # 周期性边界条件
                    up_down_i = torch.arange(1, N1 * N2 + 1)
                    up_down_j = torch.arange(N1 * N2 * (N3 - 1), totalN)
                    
                    i = torch.cat([i, up_down_i])
                    j = torch.cat([j, up_down_j])
                v = torch.ones(len(i), dtype=data_type,device=device) / 2
            else:
                raise ValueError('direction must be 1, 2, 3 or None')
        i = i - 1  # Convert to 0-based index
        Jmtx=torch.sparse_coo_tensor(torch.stack([i, j]), v, (totalN, totalN),dtype=data_type,device=device)
       
    return Jmtx+Jmtx.t()
def is_bool_or_single_bool_list(x):
    if isinstance(x, bool):
        return True
    elif isinstance(x, list) and len(x) == 1 and isinstance(x[0], bool):
        return True
    return False
def create_bool_list(x, y) -> list[bool]:
    """创建与y同长的布尔列表"""
    # 获取实际的布尔值
    if isinstance(x, bool):
        bool_val = x
    elif isinstance(x, list) and len(x) == 1 and isinstance(x[0], bool):
        bool_val = x[0]
    else:
        raise ValueError("x必须是布尔值或单元素布尔值列表")
    
    return [bool_val] * len(y)    

def ode_rk45(ode_fun: Callable, t_span: torch.Tensor, y0: torch.Tensor, 
             options: Optional[Dict[str, Any]] = None) -> Tuple[torch.Tensor, torch.Tensor, Dict, Dict]:
    """
    Modified rk45
    
    Parameters:
    -----------
    ode_fun : callable
        ODE function: f(t, y) 
    t_span : torch.Tensor
        Time span [t0, t1, ..., tfinal]
    y0 : torch.Tensor
        Initial conditions
    options : dict
        Options dictionary with keys:
        - rel_tol = 1e-3 ：relative tolerence
        - abs_tol = 1e-6 ：absolute tolerence
        - waitbar = True : whether to show progress
        - max_consecutive_failures = 10: Maximum number of consecutive step failures
    
    Returns:
    --------
    t : torch.Tensor
        Time points
    y : torch.Tensor
        Solution values
    stats : dict
        Statistics (n_calls)
    err_info : dict
        Error history and max step error
    """
    solver= 'RK45'
    sf=eq_solver(ode_fun, t_span, y0, solver, options)               
    bar=Wait_bar(t_span, sf.waitbar)  # Initialize the progress bar
    t, ang, stats, error_info =sf.run(bar)
    return t, ang, stats, error_info

def ode3_rk45(ode_fun: Callable, t_span: torch.Tensor, y0: torch.Tensor, 
             options: Optional[Dict[str, Any]] = None) -> Tuple[torch.Tensor, torch.Tensor, Dict, Dict]:
    """
    Modified rk45
    
    Parameters:
    -----------
    ode_fun : callable
        ODE function: f(t, y) 
    t_span : torch.Tensor
        Time span [t0, t1, ..., tfinal]
    y0 : torch.Tensor
        Initial conditions
    options : dict
        Options dictionary with keys:
        - rel_tol = 1e-3 ：relative tolerence
        - abs_tol = 1e-6 ：absolute tolerence
        - waitbar = True : whether to show progress
        - max_consecutive_failures = 10: Maximum number of consecutive step failures
    
    Returns:
    --------
    t : torch.Tensor
        Time points
    y : torch.Tensor
        Solution values
    stats : dict
        Statistics (n_calls)
    err_info : dict
        Error history and max step error
    """
    solver= 'RK45'
    fix_step = bool((options or {}).get('fix_step', False))
    if fix_step:
        sf = eq3_fix_solver(ode_fun, t_span, y0, solver, options)
    else:
        sf = eq3_solver(ode_fun, t_span, y0, solver, options)
    bar=Wait_bar(t_span, sf.waitbar)  # Initialize the progress bar
    t, ang, stats, error_info =sf.run(bar)
    return t, ang, stats, error_info

def ode_sde_em(f: Callable, t_span: torch.Tensor, y0: torch.Tensor, options: Optional[Dict[str, Any]] = None) -> Tuple[torch.Tensor, torch.Tensor, Dict, Dict]:
    solver = 'EM'
    sf = sde_solver(f, t_span, y0, solver, options)
    bar = Wait_bar(t_span, sf.waitbar)
    t, y, stats, err_info = sf.run(bar)
    return t, y, stats, err_info

def ode3_sde_em(f: Callable, t_span: torch.Tensor, y0: torch.Tensor, options: Optional[Dict[str, Any]] = None) -> Tuple[torch.Tensor, torch.Tensor, Dict, Dict]:
    solver = 'EM'
    fix_step = bool((options or {}).get('fix_step', False))
    if fix_step:
        sf = sde3_fix_solver(f, t_span, y0, solver, options)
    else:
        sf = sde3_solver(f, t_span, y0, solver, options)
    bar = Wait_bar(t_span, sf.waitbar)
    t, y, stats, err_info = sf.run(bar)
    return t, y, stats, err_info

def ode3_sde_eh(f: Callable, t_span: torch.Tensor, y0: torch.Tensor, options: Optional[Dict[str, Any]] = None) -> Tuple[torch.Tensor, torch.Tensor, Dict, Dict]:
    solver = 'EH'
    sf = sde3_solver(f, t_span, y0, solver, options)
    bar = Wait_bar(t_span, sf.waitbar)
    t, y, stats, err_info = sf.run(bar)
    return t, y, stats, err_info


import magbox
import numpy as np
import os
import pytest

testdata = [ # N, K, J, K_hard, fix_step, dh
    (256,1.0,0.5,1, True, 5e-2),
    (512,1.0,0.5,2, True, 5e-2),
    (1024,1.0,0.5,3, True, 5e-2),
    (256,1.0,1.0,4, True, 5e-2),
    (256,1.0,2.0,5, True, 5e-2),
    (256,1.0,3.0,5, True, 5e-2),

    (256,1.0,0.5,1, False, 5e-2),
    (512,1.0,0.5,2, False, 5e-2),
    (1024,1.0,0.5,3, False, 5e-2),
    (256,1.0,1.0,4, False, 5e-2),
    (256,1.0,2.0,5, False, 5e-2),
    (256,1.0,3.0,5, False, 5e-2),
]

def plot_fun(t,x,y,ft_abs,q,w,dispersion,dispersion_theory,err,mean_err,max_err):
    if "PYTEST_CURRENT_TEST" in os.environ: 
        print("测试模式不作图")
        return
    import matplotlib.pyplot as plt
    w_max=dispersion_theory.max()
    plt.figure(figsize=(10,10))

    plt.subplot(2,2,1)
    plt.imshow(ft_abs.T,
            origin='lower',
            extent=(q[0],q[-1],w[0],w[-1]),
    )
    plt.ylim(0,2*w_max)

    plt.subplot(2,2,2)
    plt.scatter(q,dispersion,label="Simulation")
    plt.plot(q,dispersion_theory,label="Theory")
    plt.legend()

    plt.subplot(2,2,3)
    plt.scatter(q,err,label="Error")
    plt.legend()
    plt.title(f"mean error: {mean_err:.2e}, max err: {max_err:.2e}")

    plt.subplot(2,2,4)
    plt.plot(t,x[:,0],label="x")
    plt.plot(t,y[:,0],label="y")
    plt.legend()
    # plt.title(f"mean error: {mean_err:.2e}, max err: {max_err:.2e}")
    plt.show()

@pytest.mark.parametrize("N,K,J,K_hard, fix_step, dh",testdata)
def test_spin_chain(N,K,J,K_hard,fix_step,dh):
    # N=256
    # K=1.0
    # J=0.5
    # dt=1
    # T=50
    rng=np.random.default_rng()
    theta0=np.ones(N)*0.01
    phi0=rng.random(N)*2*np.pi
    z=np.cos(theta0)
    x=np.sin(theta0)*np.cos(phi0)
    y=np.sin(theta0)*np.sin(phi0)
    # phi0=10*np.arange(N)/N *2*np.pi
    hard_kernel = K_hard * np.array([[0,0,0],[0,1,0],[0,0,0]])
    def hard_axis_heff(kernel,t, cartS):
        return -cartS @ kernel[0]
    LT = magbox.Lattice(type="square", size=[N], periodic=True)
    vars = magbox.Vars(K1=K, J=J, custom_kernel=(hard_kernel,), custom_heff=hard_axis_heff)

    def dispersion_fun(qf):
        return np.sqrt((K+J*(1-np.cos(qf))*np.cos(np.mean(theta0))+K_hard)*(K+J*(1-np.cos(qf))*np.cos(np.mean(theta0))))
    q=np.fft.fftfreq(N,1)*2*np.pi
    q=np.fft.fftshift(q)
    dispersion_theory=dispersion_fun(q)
    w_max=dispersion_theory.max()
    W_diff=np.min(np.abs(np.diff(dispersion_theory)))
    print(f"freq max: {w_max:.3e}, freq diff: {W_diff:.3e}")
    dt=np.max([2*np.pi/(4*w_max),0.05])
    T=np.min([int(3*2*np.pi/W_diff),2e4])
    print(f"use dt: {dt:.3e}, Total Time: {T:.3e}")

    sf=magbox.llg3(x,y,z,LT,vars,
                   device='cpu', dtype='f32', alpha=0,T=T,dt=dt, fix_step=fix_step, dh=dh)

    t_tc,S,stats,err_info=sf.run()
    print(S.shape)

    t=t_tc.cpu().detach().numpy()
    S=S.reshape(len(t),N,3)
    

    x=S[...,0].detach().cpu().numpy().squeeze()
    y=S[...,1].detach().cpu().numpy().squeeze()
    z=S[...,2].detach().cpu().numpy().squeeze()

    # x=S[:,::3,:].detach().cpu().numpy().squeeze()
    # y=S[:,1::3,:].detach().cpu().numpy().squeeze()
    # z=S[:,2::3,:].detach().cpu().numpy().squeeze()

    u=x+1j*y
    ft=np.fft.fft2(u.T)
    ft_abs=np.abs(ft)
    w=np.fft.fftfreq(len(t), dt)*2*np.pi
    

    ft_abs=np.fft.fftshift(ft_abs)
    w=np.fft.fftshift(w)
    

    dispersion=np.zeros(len(q))
    for idx in range(len(q)):
        arg_max=np.argmax(ft_abs[idx,:])
        dispersion[idx]=w[arg_max]
    dispersion=np.abs(dispersion)

    err=dispersion/dispersion_theory-1
    max_err=np.max(np.abs(err))
    mean_err=np.mean(np.abs(err))

    print(f"mean error: {mean_err:.2e}, max err: {max_err:.2e}")

    plot_fun(t,x,y,ft_abs,q,w,dispersion,dispersion_theory,err,mean_err,max_err)

    assert mean_err<1e-2

if __name__=="__main__":
    test_spin_chain(32,1,1,0, True, 1e-1)


# spin_chain_test(256,1,0.5,1,50)



import magbox
import numpy as np
import os
import pytest
import scipy

testdata = [ # N, K, J, alpha, fix_step, dh
    (256, 2.0, 2.0, 0.1, True, 5e-2),
    (256, 3.0, 8.0, 0.1, True, 5e-2),
    (256, 1.0, 2.0, 0.01, True, 5e-2),
    (256, 1.0, 2.0, 0.05, True, 5e-2),
    (128, 1.0, 2.0, 0.05, True, 5e-2),

    (256, 2.0, 2.0, 0.1, False, 5e-2),
    (256, 3.0, 8.0, 0.1, False, 5e-2),
    (256, 1.0, 2.0, 0.01, False, 5e-2),
    (256, 1.0, 2.0, 0.05, False, 5e-2),
    (128, 1.0, 2.0, 0.05, False, 5e-2),
]

def plot_fun(N,err,mean_err,max_err):
    if "PYTEST_CURRENT_TEST" in os.environ:
        print("测试模式不作图")
        return
    import matplotlib.pyplot as plt
    plt.figure(figsize=(10,10))

    # plt.subplot(2,2,1)
    # plt.imshow(en,
    #         origin='lower',
    #         extent=(t[0],t[-1],1,N),
    # )
    # plt.xlabel('t')
    # plt.ylabel('index')
    
    # plt.subplot(2,2,2)
    # err_tmp=err.reshape(-1)
    plt.scatter(np.linspace(1,N,N),err,label="Error")
    plt.legend()
    plt.title(f"mean error: {mean_err:.2e}, max err: {max_err:.2e}")

    # plt.subplot(2,2,3)
    # plt.legend()
    # plt.xlabel('t')
    # plt.ylabel('energy')
    # plt.title(f"index: {id},mean err: {err_t[id]:.2e}")
    plt.show()

def peak_correction(left_height,right_height,peak_height):
    if left_height > right_height:
        h1=left_height
        loc=-1
    else:
        h1=right_height
        loc=1
    epsilon=h1/(h1+peak_height)
    
    return loc*epsilon

# @pytest.skip(reason="Under rebuild")
@pytest.mark.parametrize("N,K,J,alpha, fix_step, dh",testdata)
def test_energy_damping(N,K,J,alpha, fix_step, dh):
    # N=256
    # K=1.0
    # J=0.5
    # dt=1
    # T=50
    # rng=np.random.default_rng()
    dt=2*np.pi/(K+2*J)
    loops=np.floor(1/alpha)*200
    T=np.min([np.ceil(loops*dt), 20000.0])

    dt=T/np.ceil(T/dt)
    dt=dt/8
    theta0=np.ones(N)*0.01
    phi0=np.random.rand(N)*2*np.pi

    z0=np.cos(theta0)
    x0=np.sin(theta0)*np.cos(phi0)
    y0=np.sin(theta0)*np.sin(phi0)
    
    LT = magbox.Lattice(type="square", size=[N], periodic=True)
    vars = magbox.Vars(K1=K, J=J)

    sf=magbox.llg3(x0,y0,z0, LT,vars,
                   device='cpu', dtype='f64', alpha=alpha,T=T,dt=dt,rtol=1e-4, fix_step=fix_step, dh=dh)

    t_tc,S,*_=sf.run()
    t=t_tc.cpu().detach().numpy()
    S=S.reshape(len(t),N,3)

    x=S[...,0].detach().cpu().numpy()
    y=S[...,1].detach().cpu().numpy()
    z=S[...,2].detach().cpu().numpy()

    u=x+1j*y
    ft=np.fft.fft2(u.T)
    ft_abs=np.abs(ft)
    w=np.fft.fftfreq(len(t), dt)*2*np.pi
    q=np.fft.fftfreq(N,1)*2*np.pi

    ft_abs=np.fft.fftshift(ft_abs)
    w=np.fft.fftshift(w)
    dw=w[1]-w[0]
    q=np.fft.fftshift(q)

    alpha_q=np.zeros(len(q))
    alpha_qc=np.zeros(len(q))
    for idx in range(len(q)):
        peaks,property=scipy.signal.find_peaks(ft_abs[idx,:]**2,width=(None,None))
        peak_idx=np.argmax(property['prominences'])
        omega=w[peaks[peak_idx]]
        omega_correction=peak_correction(ft_abs[idx,peaks[peak_idx]-1],ft_abs[idx,peaks[peak_idx]+1],ft_abs[idx,peaks[peak_idx]])*dw
        width=property['widths'][peak_idx]*dw
        alpha_q[idx]=width/omega/2
        alpha_qc[idx]=width/2/(omega-omega_correction)
    err=alpha_q/alpha-1
    mean_err=np.mean(np.abs(err))
    max_err=np.max(np.abs(err))

    err_c=alpha_qc/alpha-1
    mean_errc=np.mean(np.abs(err_c))
    max_errc=np.max(np.abs(err_c))
    

    print(f"mean error: {mean_err:.2e}, max err: {max_err:.2e}")
    print(f"mean error correction: {mean_errc:.2e}, max err correction: {max_errc:.2e}")
    print(f"Technical accuracy: {2*dw/(alpha*K):.2e}")

    plot_fun(N,err,mean_err,max_err)

    assert mean_err<2*dw/(alpha*K)

if __name__=="__main__":
    test_energy_damping(32, 1.0, 2.0, 0.001, True, 2e-1)


# spin_chain_test(256,1,0.5,1,50)



from magbox.Wait_bar import Wait_bar
import torch
import math
import warnings
from .eq_solver import eq_solver

class sde_solver(eq_solver):
    def _tableau(self, solver_name):
        device=self.device
        dtype=self.dtype
        if solver_name == 'EM':
            self.order=2
            self.adaptive_order = 1.5
            sqrt2=torch.tensor(math.sqrt(2),dtype=dtype,device=device)
            self.alpha=torch.tensor([],
                                    dtype=dtype,device=device)
            self.beta=torch.tensor([],
                                   dtype=dtype,device=device)
            self.c_error=[torch.tensor([1/2, -1/2],dtype=dtype,device=device).view(-1,1),
                        torch.tensor([-sqrt2/2, -sqrt2/2, sqrt2/2],dtype=dtype,device=device).view(-1,1)]
            self.interp_coeff=torch.tensor([[-3,2],
                                            [4,-4],
                                            [-1,2]],
                                    dtype=dtype,device=device)  
            pass
        else:
            super()._tableau(solver_name)
    @staticmethod
    def _one_step(f1, g1, t, y, h, W, sde_fcn, order, y_list,f_list,gw_list):
        # EM method one step
        h_abs = torch.abs(h)
        # Drift uses signed h, diffusion uses |h|.
        half_h = h * 0.5
        sqrt_half_abs_h = torch.sqrt(h_abs * 0.5)

        # stage 1
        y_list[0] = y
        f_list[0] = f1
        gw11 = g1(W[0])
        gw_list[0] = gw11

        y2 = y + f1 * half_h + gw11 * sqrt_half_abs_h
        y_list[1] = y2
        t2 = t + half_h
        f2, g2, _ = sde_fcn(t2, y2)
        f_list[1] = f2

        # stage 2
        gw22 = g2(W[1])
        gw_list[1] = gw22
        y_new = y2 + f2 * half_h + gw22 * sqrt_half_abs_h
        y_list[2] = y_new
        t_new = t + h

        # combined noise term for error estimate
        gw_list[2] = g1(W[0] + W[1])
        return t_new, y_list, f_list, gw_list
    @staticmethod
    def _get_err(f_list, gw_list, c_error, h, y, y_new, atol):
        err = h * torch.einsum('s...,si->...', f_list, c_error[0])
        err += torch.sqrt(h) * torch.einsum('s...,si->...', gw_list, c_error[1])
        err /= torch.max(torch.max(y.abs(), y_new.abs()), atol)
        err = torch.max(err.abs())
        return err
    
    @staticmethod
    def _interp_fun(t_interp, t, y, h, y_list, interp_coeff):
        y_interp = eq_solver._interp_fun(t_interp, t, y, h, y_list, interp_coeff,is_dy=False)
        return y_interp
    
    def run(self, bar: Wait_bar):
        t_out, y_out, t, y, dtype, device, h_max, h_min, rtol, atol, max_failures, t_final, t0, t_dir, t_span, waitbar, alpha, beta, c_error, sde_fcn, S, chunk, output_pos, n_t_span, n_eq, refine, order, interp_coeff = self._get_parameters()
        if hasattr(self, 'adaptive_order'):
            adaptive_order = self.adaptive_order
        else:
            adaptive_order = order - 0.5

        n_out = 0
        t_out[n_out] = t
        y_out[n_out, ...] = y

        error_history = []
        n_calls = 0
        n_steps = 0
        
        h = torch.min(h_max, torch.max(h_min, 0.1 * torch.abs(t_final - t0)))
        h_abs = torch.abs(h)

        f1, g1, noise_dim = sde_fcn(t, y)
        n_calls += 1

        # Pre-allocate buffers for EM stages to reduce per-step allocations.
        # y_list: (order+1, ...)  f_list: (order, ...)  gw_list: (order+1, ...)
        y_list = torch.empty((order + 1,) + y.shape, dtype=dtype, device=device)
        f_list = torch.empty((order,) + y.shape, dtype=dtype, device=device)
        gw_list = torch.empty((order + 1,) + y.shape, dtype=dtype, device=device)

        finished = False
        next_idx = 1
        n_failures = 0
        integration_failed = False

        while not finished:
            h_abs = torch.min(h_max, torch.max(h_min, h_abs))
            h = t_dir * h_abs
            if 1.1 * h_abs >= torch.abs(t_final - t):
                h = t_final - t
                h_abs = torch.abs(h)
                finished = True
            
            no_failed = True

            W=torch.randn((order,)+noise_dim,dtype=dtype,device=device)

            while True:
                t_new, y_list, f_list, gw_list = self._one_step(f1, g1, t, y, h, W, sde_fcn, order, y_list, f_list, gw_list)
                
                n_calls += 1
                y_new= y_list[order]
                
                err = self._get_err(f_list, gw_list, c_error, h_abs, y, y_new, atol)
                accept_step = err <= rtol

                if accept_step:
                    n_failures = 0
                
                if h_abs <= h_min:
                    accept_step = True
                    n_failures += 1
                    no_failed = False
                    if n_failures >= max_failures:
                        bar.close(waitbar)
                        warnings.warn(
                            f"Step size reached minimum hmin = {h_min.item():.2e} at t={t.item():.2e}, but still cannot satisfy tolerance. "
                            f"Current error: {err:.2e}, Required tolerance: {rtol:.2e}. "
                            f"This may indicate a stiff ODE or overly strict tolerances. "
                            f"Consider using a stiff solver or relaxing tolerances.",
                            RuntimeWarning,
                            stacklevel=2
                        )
                        finished = True
                        integration_failed = True
                        break
                else:
                    n_failures = 0
                
                if accept_step:
                    error_history.append(err)
                    break
                else:
                    if no_failed:
                        no_failed = False
                        h_abs = torch.max(h_min, h_abs * max(0.1, 0.8 * (rtol / err) ** (1/adaptive_order)))
                    else:
                        h_abs = torch.max(h_min, 0.5 * h_abs)
                    h = t_dir * h_abs
                    finished = False
            
            n_steps += 1
            if integration_failed:
                break
            
            bar.update(t_new, h, waitbar, finished)

            # Output processing using ntrp
            n_out_new, t_out_new, y_out_new, next_idx = self._get_outputs(output_pos, self._interp_fun, t, S, refine, t_new, y_new, y, h, y_list, interp_coeff, dtype, device, n_t_span, t_span, next_idx, t_dir)
            y_out_new = self._after_process(y_out_new)
            
            if n_out_new > 0:
                old_n_out = n_out
                n_out = n_out + n_out_new
                
                if n_out + 1 > t_out.shape[0]:
                    extra = max(chunk, n_out_new)
                    t_out_new_temp = torch.zeros(t_out.shape[0] + extra, dtype=dtype, device=device)
                    t_out_new_temp[:t_out.shape[0]] = t_out
                    t_out = t_out_new_temp
                    
                    y_out_new_temp = torch.zeros((y_out.shape[0] + extra,)+n_eq, dtype=dtype, device=device)
                    y_out_new_temp[:y_out.shape[0],...] = y_out
                    y_out = y_out_new_temp
                
                t_out[old_n_out+1:n_out+1] = t_out_new
                y_out[old_n_out+1:n_out+1,...] = y_out_new

            h_abs = self._optimal_step_size(h_abs, err/rtol, adaptive_order, no_failed)
            
            t = t_new
            y = y_new

            y = self._after_process(y) 

            f1, g1, _ = sde_fcn(t, y)
            n_calls += 1

        bar.close(waitbar)
        
        t_out = t_out[:n_out+1]
        y_out = y_out[:n_out+1, ...]

        stats = {'n_calls': n_calls,
             'n_steps': n_steps,
             'n_output': n_out+1,
             'intergration': not integration_failed}
        err_info = {
            'err_history': error_history,
            'max_step_error': max(error_history) if error_history else 0.0
        }
        return t_out, y_out, stats, err_info


# ==========  1. 生成数据  ==========
from __future__ import annotations
import numpy as np
from pathlib import Path
from magbox import llg3, spin3, Lattice, Vars
import matplotlib.pyplot as plt
from torch import sin, cos

def run_thermal(dt: float = 0.1,
                alpha: float =0.1,
                Temp: float= 0.1,
                gamma: float =1,
                T: float = 100,
                run_id: int =1,
                out_dir: Path  = Path("."),
                device: str = "cpu",
                dtype: str = 'f32',
                seed = None,
                spin_num: int = 2**8,
                rtol: float = 1e-2,
                atol: float = 1e-4,
                fix_step: bool = False,
                dh: float | None = None) -> Path:
    """
    运行一次 LLG 热模拟，把结果保存到 <out_dir>/thermal_test_<run_id>.npz
    文件中额外存下所有输入参数，方便后期读取。
    """
    out_dir = Path(out_dir)
    out_dir.mkdir(exist_ok=True)

    rng = np.random.default_rng(seed)
    N1 = spin_num
    lt = Lattice(type="square", size=[N1], periodic=True)
    vars = Vars(J=0)

    theta0 = rng.random(N1) * 0.1
    phi0 = rng.random(N1) * 2 * np.pi

    x0=np.sin(theta0)*np.cos(phi0)
    y0=np.sin(theta0)*np.sin(phi0)
    z0=np.cos(theta0)

    sf = llg3(x0,y0,z0, lt, vars=vars, device=device, dtype=dtype,
              dt=dt, alpha=alpha, T = T, Temp=Temp, gamma=gamma, rtol=rtol, atol=atol, fix_step=fix_step, dh=dh)
    t, S, stats, erro_info = sf.run()
    S=S.reshape(len(t), N1, 3)
    z=S[..., 2].detach().cpu().numpy()

    en=1/2*(1-z**2)

    t_np = t.cpu().detach().numpy()

    save_file = out_dir / f"thermal_test_{run_id}.npz"
    np.savez(save_file,
             t_np=t_np,
             en=en,
             alpha=alpha,
             Temp=Temp,
             dt=dt,
             gamma=gamma,
             T=T,)
    print(f"[run_thermal] 数据已保存 -> {save_file}")
    return save_file
def density(e):
    eps=np.finfo(float).eps
    e=np.clip(e, eps, 0.5-eps)

    se=np.sqrt(1-2*e)
    return -se

# ==========  2. 读取并画图  ==========
def plot_thermal(npz_file:  Path,
                 start_time: float = 20,
                 bins: int = 100,
                 fit_counts=None,
                 fig_dir: Path = Path(".")):
    """
    读取 run_thermal 生成的 npz 文件，
    画能量时序图（并标出 start_time 位置）以及能量直方图 + 线性拟合。
    图片保存为与 npz 同名（扩展名换成 .png）。
    """
    npz_file = Path(npz_file)
    fig_dir = Path(fig_dir)
    fig_dir.mkdir(exist_ok=True)

    data = np.load(npz_file)
    t_np = data['t_np']
    en = data['en']
    alpha = data['alpha']
    Temp = data['Temp']
    dt = data['dt']
    gamma = data['gamma']

    # 计算截取起点对应的索引
    idx = min(len(t_np) -2*bins, int(start_time / (t_np[1] - t_np[0])))

    # 直方图
    en_mean=np.mean(en,axis=1)
    en_st=en[idx:,:]
    en_st=en_st.reshape(-1)
    hist, bin_edges = np.histogram(en_st, bins=bins)
    bin_centers = 0.5 * (bin_edges[:-1] + bin_edges[1:])
    # 态密度
    ds=density(bin_edges)
    dens=ds[1:]-ds[:-1]
    dens/=np.diff(bin_edges).mean()

    loghist=np.log((hist+1)/dens)

    # 拟合
    if fit_counts is None:
        fit_counts = len(bin_centers)
    # (k,b), cov = np.polyfit(bin_centers[:fit_counts], loghist[:fit_counts], deg=1,cov=True)
    (k,b), cov = np.polyfit(bin_centers[:fit_counts], loghist[:fit_counts], deg=1,cov=True, w= np.sqrt(hist[:fit_counts]+1))
    x_smooth = np.linspace(bin_centers[:fit_counts].min(), bin_centers[:fit_counts].max(), 200)
    k_err, b_err=np.sqrt(np.diag(cov))
    y_smooth = k * x_smooth + b

    # 画图
    fig, axs = plt.subplots(2, 1, figsize=(15, 10))

    # 图 1：能量时序
    axs[0].plot(t_np, en_mean, label="energy")
    axs[0].axvline(t_np[idx], color='red', linestyle='--', label=f'start={start_time}')
    axs[0].set_xlabel("Time")
    axs[0].set_title("energy")
    axs[0].legend()

    # 图 2：散点图 + 拟合
    axs[1].scatter(bin_centers, loghist,
                   s=60, facecolors='none', edgecolors='dodgerblue', lw=1.2,
                   alpha=0.85, label='Energy counts')
    axs[1].plot(x_smooth, y_smooth, color='red',
                label=f'Fit y=({k:.2f}±{k_err:.2f})x+({b:.2f}±{b_err:.2f})')
    axs[1].set_xlabel("Energy")
    axs[1].set_ylabel("Counts")
    axs[1].set_title(f"Energy Histogram "
                     f"(file={npz_file.stem}, α={alpha}, 1/T={1/Temp:.2e}, "
                     f"dt={dt}, γ={gamma}), total number={np.sum(hist):.1e}")
    axs[1].legend()

    fig.tight_layout()

    # 保存
    fig_file = fig_dir / (npz_file.stem + ".png")
    fig.savefig(fig_file, dpi=300)
    print(f"[plot_thermal] 图片已保存 -> {fig_file}")
    plt.close(fig)
    return fig_file


# ==========  3. 示例用法  ==========
if __name__ == "__main__":
    # 生成一组数据
    npz = run_thermal(dt=0.1, alpha=0.2, Temp=0.5,
                      gamma=0.17, T=100, run_id=1)

    # 读取并画图
    plot_thermal(npz, start_time=30, bins=60)


import numpy as np
import torch
from . import boxlib
from .initial import Lattice

class spin3:
    def __init__(self, x, y, z, lattice_type: Lattice,dtype="f32", device="gpu",thread:int=4, require_ini_grad:bool=False):
        torch.set_num_threads(thread)

        dtype=boxlib.get_data_type(dtype)
        self.dtype=dtype
        self.device=boxlib.get_device(device)
        self.require_ini_grad=require_ini_grad

        self.x = torch.as_tensor(x,dtype=self.dtype,device=self.device)
        self.y = torch.as_tensor(y,dtype=self.dtype,device=self.device)
        self.z = torch.as_tensor(z,dtype=self.dtype,device=self.device)

        if self.require_ini_grad:
            self.x.requires_grad_(True)
            self.y.requires_grad_(True)
            self.z.requires_grad_(True)
        
        self.x=self.x.view(-1,1)
        self.y=self.y.view(-1,1)
        self.z=self.z.view(-1,1)
        l_x=len(self.x)
        l_y=len(self.y)
        l_z=len(self.z)
        assert l_x==l_y and l_y==l_z, "mx, my and mz must have the same length"
        self.num = l_x

        self.update()

        self.lattice_type=lattice_type

        num=lattice_type.size
        assert self.num == np.prod(num), f"initial condition length {self.num} does not match lattice size {num}"
    def update(self): # 根据x,y,z的值重新归一化并更新cart_S的值
        self.cart_S=torch.cat([self.x,self.y,self.z], dim = 1)
        self.cart_S=torch.nn.functional.normalize(self.cart_S, dim=1,p=2)
        self.x, self.y ,self.z = spin3.get_xyz(self.cart_S)
        self.S=self.cart_S.view(-1,1)
    @staticmethod
    def get_xyz(S):
        return S[::3],S[1::3],S[2::3]
        



import torch

from magbox.Wait_bar import Wait_bar
from magbox.solver.sde3_solver import sde3_solver


class sde3_fix_solver(sde3_solver):
    """Fixed-step SDE solver for 3D spin dynamics.

    Uses the same EM stepping kernel as `sde_solver`, but advances with a fixed step `dh`.
    `dh` is read from `options['dh']` and defaults to `(t_span[1]-t_span[0]) / 10` when possible.
    """

    def _ode_options(self, options):
        super()._ode_options(options)
        if options is None:
            options = {}
        dh = options.get('dh', None)
        if dh is None:
            self.dh = None
        else:
            self.dh = torch.as_tensor(dh, device=self.device, dtype=self.dtype)
    @staticmethod
    def _one_step(f1, g1, t, y, h, W, sde_fcn, order, y_list,f_list,gw_list):
        # EM method one step
        h_abs = torch.abs(h)
        # Drift uses signed h, diffusion uses |h|.
        half_h = h * 0.5
        sqrt_half_abs_h = torch.sqrt(h_abs * 0.5)

        # stage 1
        y_list[0] = y
        f_list[0] = f1
        gw11 = g1(W[0])
        gw_list[0] = gw11

        y2 = y + f1 * half_h + gw11 * sqrt_half_abs_h
        y_list[1] = y2
        t2 = t + half_h
        f2, g2, _ = sde_fcn(t2, y2)
        f_list[1] = f2

        # stage 2
        gw22 = g2(W[1])
        gw_list[1] = gw22
        y_new = y2 + f2 * half_h + gw22 * sqrt_half_abs_h
        y_list[2] = y_new
        t_new = t + h

        ## combined noise term for error estimate
        ## not used in fixed step mode
        # gw_list[2] = g1(W[0] + W[1])
        return t_new, y_list, f_list, gw_list

    def run(self, bar: Wait_bar):
        (
            t_out,
            y_out,
            t,
            y,
            dtype,
            device,
            h_max,
            h_min,
            rtol,
            atol,
            max_failures,
            t_final,
            t0,
            t_dir,
            t_span,
            waitbar,
            alpha,
            beta,
            c_error,
            sde_fcn,
            S,
            chunk,
            output_pos,
            n_t_span,
            n_eq,
            refine,
            order,
            interp_coeff,
        ) = self._get_parameters()

        # Fixed step size (magnitude), clamped to [h_min, h_max]
        if self.dh is None:
            if isinstance(t_span, torch.Tensor) and t_span.numel() >= 2:
                dt = torch.abs(t_span[1] - t_span[0])
            else:
                dt = torch.abs(t_final - t0)
            dh_abs = dt / 10
        else:
            dh_abs = torch.abs(self.dh)
        if dh_abs.numel() != 1:
            dh_abs = dh_abs.reshape(())
        dh_abs = torch.min(h_max, torch.max(h_min, dh_abs))
        dh_abs = torch.as_tensor(dh_abs, dtype=dtype, device=device)

        # Outputs
        n_out = 0
        t_out[n_out] = t
        y_out[n_out, ...] = self._after_process(y)

        # Output pointer for t_span output
        next_idx = 1

        # Initial drift/diffusion
        f1, g1, noise_dim = sde_fcn(t, y)
        n_calls = 1
        n_steps = 0

        # Pre-allocate buffers for EM stages
        # y_list: (order+1, ...)  f_list: (order, ...)  gw_list: (order+1, ...)
        y_list = torch.empty((order + 1,) + y.shape, dtype=dtype, device=device)
        f_list = torch.empty((order,) + y.shape, dtype=dtype, device=device)
        gw_list = torch.empty((order + 1,) + y.shape, dtype=dtype, device=device)

        err_history: list[float] = []

        finished = False
        while not finished:
            if dh_abs > torch.abs(t_final-t):
                dh_abs = torch.abs(t_final-t)
                finished = True

            h = t_dir * dh_abs

            # One fixed EM step
            W = torch.randn((order,) + noise_dim, dtype=dtype, device=device)
            t_new, y_list, f_list, gw_list = self._one_step(
                f1, g1, t, y, h, W, sde_fcn, order, y_list, f_list, gw_list
            )
            n_calls += 1
            y_new = y_list[order]

            finished = bool(t_dir * (t_final - t_new) <= 0)
            bar.update(t_new, h, waitbar, finished)

            # outputs
            nout_new, t_out_new, y_out_new, next_idx = self._get_outputs(output_pos,self._interp_fun, t, S, refine, t_new, y_new, y, h, y_list, interp_coeff, dtype, device, n_t_span, t_span, next_idx,t_dir)
            y_out_new = self._after_process(y_out_new)

            if nout_new > 0:
                old_nout = n_out
                n_out += nout_new

                if n_out + 1 > t_out.shape[0]:
                    extra = max(chunk, nout_new)
                    tout_new_temp = torch.zeros(t_out.shape[0] + extra, dtype=dtype, device=device)
                    tout_new_temp[: t_out.shape[0]] = t_out
                    t_out = tout_new_temp

                    yout_new_temp = torch.zeros((y_out.shape[0] + extra,) + n_eq, dtype=dtype, device=device)
                    yout_new_temp[: y_out.shape[0], ...] = y_out
                    y_out = yout_new_temp

                t_out[old_nout + 1 : n_out + 1] = t_out_new
                y_out[old_nout + 1 : n_out + 1, ...] = y_out_new

            # Advance
            n_steps += 1
            t = t_new
            y = self._after_process(y_new)

            f1, g1, _ = sde_fcn(t, y)
            n_calls += 1

        bar.close(waitbar)

        if output_pos == 1:
            if n_t_span > 0:
                final_len = max(1, min(next_idx, n_t_span))
                t_out = t_out[:final_len]
                y_out = y_out[:final_len, ...]
        else:
            t_out = t_out[: n_out + 1]
            y_out = y_out[: n_out + 1, ...]

        stats = {
            'n_calls': n_calls,
            'n_steps': n_steps,
            'n_output': y_out.shape[0],
            'integration': True,
            'fix_step': True,
            'dh': float(dh_abs.item()),
        }
        err_info = {
            'err_history': err_history,
            'max_step_error': 0.0,
        }
        return t_out, y_out, stats, err_info



from magbox.solver.eq_solver import eq_solver
from .vec_normalize import vec_normalize
import torch

class eq3_solver(eq_solver):
    def _after_process(self, y: torch.Tensor) -> torch.Tensor:
        return vec_normalize(y)


from tqdm import tqdm
import time
import sys

class Wait_bar:
    def __init__(self, t_span, flag):
        self.t0=0
        if flag:
            h=0
            self.t0 = t_span[0].item()
            self.t_final = t_span[-1].item()
            self.total_progress = self.t_final - self.t0
            
            # Check if running in a TTY (terminal)
            is_tty = sys.stdout.isatty()
            # General purpose setting:
            # TTY: 0.2s for smooth animation.
            # Non-TTY (Slurm/Log): 5.0s. 
            #   - Short task (20s): ~4-5 updates (sufficient heartbeat).
            #   - Long task (24h): ~1.7MB log (very manageable).
            self.update_interval = 0.2 if is_tty else 5.0
            
            self.pbar = tqdm(total=self.total_progress, desc='ODE Integration', 
                    unit='time', ncols=100, bar_format='{l_bar}{bar}| {n:.2f}/{total_fmt} [{elapsed}<{remaining}{postfix}]')
            self.pbar.set_postfix(dt=format(h,'.2e'))
            self.last_update_time = time.time()
    def update(self, t, h, flag, done):
        if flag:
            current_time = time.time()
            if current_time - self.last_update_time >= self.update_interval or done:
                progress = t.item() - self.t0
                self.pbar.n = min(progress,self.total_progress)
                self.pbar.set_postfix(dt=format(h,'.2e'), refresh=False)
                self.pbar.refresh()
                self.last_update_time = current_time
    def close(self,flag):
        if flag: 
            self.pbar.close()


from __future__ import annotations
import torch
import scipy.sparse
import numpy as np
from typing import Optional, Callable, cast
from . import boxlib
from .initial import Vars, Lattice
from .spin3 import spin3
from dataclasses import replace

class heff3:
    def __init__(self, sp:spin3, vars:Vars): # B=0,B_dir=[0,0,1],J=1,K_dir=[0,0,1],K1=1,D=0
        # self.spin = spin

        B=vars.B # external field
        B_dir=vars.B_dir # external field direction

        J=vars.J # exchange interaction
        K1_dir=vars.K1_dir # uniaxial anisotropy axis
        K1=vars.K1 # uniaxial anisotropy strength
        D=vars.D # Dzyaloshinskii-Moriya interaction

        self.num=sp.num
        self.B=torch.tensor(B,dtype=sp.dtype,device=sp.device)# external field
        self.B_dir=torch.tensor(B_dir,dtype=sp.dtype,device=sp.device)
        self.B_dir=self.B_dir.view(3)
        self.B_dir=self.B_dir/torch.norm(self.B_dir)
        self.B=self.B*self.B_dir

        self.K_dir=torch.tensor(K1_dir,dtype=sp.dtype,device=sp.device)
        self.K_dir=self.K_dir.view(3)
        self.K_dir=self.K_dir/torch.norm(self.K_dir)
        self.K1=torch.tensor(K1,dtype=sp.dtype,device=sp.device)
        self.K1=self.K1*self.K_dir

        # self.H = torch.zeros((self.num,3))
        self.data_type=sp.dtype
        self.device=sp.device

        self.custom_field_fn = vars.custom_heff
        
        raw_kernel = vars.custom_kernel
        if isinstance(raw_kernel, tuple):
            processed_kernel = []
            for item in raw_kernel:
                if isinstance(item, torch.Tensor):
                    processed_kernel.append(item)
                elif scipy.sparse.issparse(item):
                    coo = item.tocoo()
                    values = coo.data
                    indices = np.array([coo.row, coo.col])
                    i = torch.tensor(indices, dtype=torch.long, device=self.device)
                    sparse_values = torch.tensor(values, dtype=self.data_type, device=self.device)
                    t = torch.sparse_coo_tensor(i, sparse_values, coo.shape, device=self.device)
                    processed_kernel.append(t)
                else:
                    t = torch.as_tensor(item, device=self.device,dtype=self.data_type)
                    if t.numel() > 100:
                        sparsity = 1.0 - (float(t.count_nonzero()) / t.numel())
                        if sparsity >= 0.5:
                            t = t.to_sparse()
                    processed_kernel.append(t)
            self.custom_kernel = tuple(processed_kernel)
        else:
            self.custom_kernel = raw_kernel

        if self.custom_field_fn is not None and not callable(self.custom_field_fn):
            raise TypeError("Vars.custom_heff must be callable or None")
        l_type=sp.lattice_type
        N_dim=l_type.size
        N_dim=len(N_dim)

        if isinstance(J,(int,float)) or (isinstance(J,list) and len(J)==1):
            J_val=J[0] if isinstance(J,list) else J
            self.has_J = bool(J_val)
            J_tensor = torch.tensor(J_val,dtype=sp.dtype,device=sp.device)
            self.Jmtx=J_tensor*boxlib.get_Jmtx(replace(l_type,J_direction=None), device=sp.device,data_type=sp.dtype)
        elif isinstance(J,list):
            if len(J)!=N_dim:
                raise ValueError(f"J should be a scalar or a list of length {N_dim}")
            # self.Jmtx=J.to(device=spin.device,dtype=spin.data_type)
            self.Jmtx=torch.zeros((self.num,self.num),dtype=sp.dtype,device=sp.device)
            for i in range(N_dim):
                J_val = J[i]
                J_tensor = torch.tensor(J_val, dtype=sp.dtype, device=sp.device)
                self.Jmtx = self.Jmtx + J_tensor*boxlib.get_Jmtx(replace(l_type,J_direction=i), device=sp.device,data_type=sp.dtype)
            self.has_J = any(bool(v) for v in J)
             
        self.has_B = bool(torch.norm(self.B).item())
        if self.has_B:
            self.B_expanded = self.B.expand(self.num, 3)
            
        self.has_K1 = bool(torch.norm(self.K1).item())
        self.has_custom = self.custom_field_fn is not None

        if self.has_custom:
            dummy_cartS = torch.ones((self.num,3), dtype=self.data_type, device=self.device)
            dummy_cartS = torch.nn.functional.normalize(dummy_cartS, dim=1,p=2)
            try:
                t=torch.tensor(0.0,dtype=self.data_type,device=self.device)
                custom_out = self.custom_field_fn(self.custom_kernel, t, dummy_cartS)
            except Exception as exc:
                raise ValueError("custom_heff raised an exception during shape validation") from exc

            if not isinstance(custom_out, torch.Tensor):
                raise TypeError("custom_heff must return a torch.Tensor")
            if custom_out.shape != (self.num,3):
                raise ValueError(f"custom_heff output shape should be {(self.num,3)}, got {tuple(custom_out.shape)}")

        if vars.require_vars_grad:
            self.B.requires_grad_(True)
            self.K1.requires_grad_(True)
            self.Jmtx.requires_grad_(True)
    def energy(self, S: torch.Tensor) -> torch.Tensor:
        x = S[::3, : ]
        y = S[::3, : ]
        z = S[::3, : ]
        t_num=x.shape[1]

        cartS=self.get_cart_S(x,y,z)

        E_total = torch.zeros(t_num, dtype=self.data_type, device=self.device)

        if self.has_B:
            E_zeeman= - torch.sum(self.B[0] * x + self.B[1] * y + self.B[2] * z, dim=0)
            E_total = E_total + E_zeeman

        if self.has_J:
            E_exch= -0.5 * torch.sum(cartS * (self.Jmtx @ cartS), dim=0)
            E_exch=E_exch[0::3] + E_exch[1::3] + E_exch[2::3]
            E_total = E_total + E_exch

        if self.has_K1:
            E_anis= -0.5 * torch.norm(self.K1) * torch.sum((x *self.K_dir[0] + y * self.K_dir[1] + z *self.K_dir[2])**2, dim=0)
            E_total = E_total + E_anis

        return E_total
    def zeeman3(self) -> torch.Tensor:
        return self.B_expanded
    # def zeeman2(self,ctheta,stheta,cphi,sphi):
    #     h_theta=self.B_dir@torch.cat([])
    # return self.B*torch.cat([*],1)

    def dipole3(self) -> torch.Tensor:
        # calculate dipole field
        return torch.zeros((self.num,3), dtype=self.data_type, device=self.device)

    def exchange3(self, cartS: torch.Tensor) -> torch.Tensor:
        # calculate exchange field
        return self.Jmtx @ cartS

    def uni_anisotropy3(self, cartS: torch.Tensor) -> torch.Tensor:
        # calculate anisotropy field
        return self.K1 * cartS

    def DM3(self) -> torch.Tensor:
        # calculate DM field
        return torch.zeros((self.num,3), dtype=self.data_type, device=self.device)

    def custom3(self,t, cartS: torch.Tensor) -> torch.Tensor:
        # user provided effective field
        if self.custom_field_fn is None:
            raise RuntimeError("custom_heff is not set; call `custom3` only when `vars.custom_heff` is provided")
        return self.custom_field_fn(self.custom_kernel, t, cartS)

    def all3(self, t: float, x: torch.Tensor, y: torch.Tensor, z: torch.Tensor) -> torch.Tensor:
        cartS=self.get_cart_S(x,y,z)
        h0 = torch.zeros((self.num,3), dtype=self.data_type, device=self.device)
        if self.has_B:
            h0.add_(self.zeeman3())
        if self.has_J:
            h0.add_(self.exchange3(cartS))
        if self.has_K1:
            h0.add_(self.uni_anisotropy3(cartS))
        if self.has_custom:
            h0.add_(self.custom3(t, cartS))

        return h0.reshape(-1,1)
    
    @staticmethod
    def get_cart_S(x: torch.Tensor, y: torch.Tensor, z: torch.Tensor) -> torch.Tensor:
        return torch.cat([x, y, z], dim=1)
    
        


import torch

from magbox.Wait_bar import Wait_bar
from magbox.solver.eq3_solver import eq3_solver


class eq3_fix_solver(eq3_solver):
    """Fixed-step RK solver for 3D spin ODEs.

    Notes
    -----
    - Uses the same Dormand–Prince (RK45) tableau as `eq_solver`, but advances with a fixed step `dh`.
    - `dh` is read from `options['dh']` and defaults to `(t_span[1]-t_span[0]) / 10` when possible.
    """

    def _ode_options(self, options):
        super()._ode_options(options)
        # NOTE: eq_solver.__init__ calls _ode_options() before _ode_initial(), so `t_span` is
        # not available here. We only record the user-provided `dh` and compute defaults later.
        if options is None:
            options = {}
        dh = options.get('dh', None)
        if dh is None:
            self.dh = None
        else:
            self.dh = torch.as_tensor(dh, device=self.device, dtype=self.dtype)

    def run(self, bar: Wait_bar):
        (
            t_out,
            y_out,
            t,
            y,
            dtype,
            device,
            h_max,
            h_min,
            rtol,
            atol,
            max_failures,
            t_final,
            t0,
            t_dir,
            t_span,
            waitbar,
            alpha,
            beta,
            c_error,
            ode_fcn,
            S,
            chunk,
            output_pos,
            n_t_span,
            n_eq,
            refine,
            order,
            interp_coeff,
        ) = self._get_parameters()

        # Fixed step size (magnitude), clamped to [h_min, h_max]
        if self.dh is None:
            if isinstance(t_span, torch.Tensor) and t_span.numel() >= 2:
                dt = torch.abs(t_span[1] - t_span[0])
            else:
                dt = torch.abs(t_final - t0)
            dh_abs = dt / 10
        else:
            dh_abs = torch.abs(self.dh)
        if dh_abs.numel() != 1:
            dh_abs = dh_abs.reshape(())
        dh_abs = torch.min(h_max, torch.max(h_min, dh_abs))

        n_beta = len(beta)

        # Pre-allocate RK45 stage buffers (7 stages).
        stage_shape = (n_beta + 1,) + y.shape
        f_list_buffer = torch.zeros(stage_shape, dtype=dtype, device=device)
        y_list_buffer = torch.zeros(stage_shape, dtype=dtype, device=device)

        # Initial derivative
        f1 = ode_fcn(t, y)
        n_calls = 1

        n_steps = 0
        # Keep outputs consistent with the post-processing convention.
        n_out = 0
        t_out[n_out] = t
        y_out[n_out, ...] = self._after_process(y)

        # Output pointer for t_span mode
        next_idx = 1

        err_history: list[float] = []

        finished = False

        # Main loop
        while not finished:
            if dh_abs > torch.abs(t_final-t):
                dh_abs = torch.abs(t_final-t)
                finished = True

            h = t_dir * dh_abs

            # One fixed step
            y_list_buffer[0] = y
            f_list_buffer[0] = f1
            t_new, y_list_buffer, f_list_buffer = self._one_step(
                f_list_buffer, y_list_buffer, t, y, h, ode_fcn, alpha, beta
            )
            n_calls += n_beta

            y_new = y_list_buffer[n_beta]

            # Update progress bar
            finished = bool(t_dir * (t_final - t_new) <= 0)
            bar.update(t_new, h, waitbar, finished)

            # outputs
            nout_new, t_out_new, y_out_new, next_idx = self._get_outputs(output_pos,self._interp_fun, t, S, refine, t_new, y_new, y, h, f_list_buffer, interp_coeff, dtype, device, n_t_span, t_span, next_idx,t_dir)
            y_out_new = self._after_process(y_out_new)
            # Store output
            if nout_new > 0:
                old_nout = n_out
                n_out += nout_new

                if n_out + 1 > t_out.shape[0]:
                    extra = max(chunk, nout_new)
                    tout_new_temp = torch.zeros(t_out.shape[0] + extra, dtype=dtype, device=device)
                    tout_new_temp[: t_out.shape[0]] = t_out
                    t_out = tout_new_temp

                    yout_new_temp = torch.zeros((y_out.shape[0] + extra,) + n_eq, dtype=dtype, device=device)
                    yout_new_temp[: y_out.shape[0], ...] = y_out
                    y_out = yout_new_temp

                t_out[old_nout + 1 : n_out + 1] = t_out_new
                y_out[old_nout + 1 : n_out + 1, ...] = y_out_new

            # Advance state
            n_steps += 1
            t = t_new
            y = self._after_process(y_new)

            # After post-processing (normalization), recompute derivative.
            f1 = f_list_buffer[order+1]

        bar.close(waitbar)

        stats = {
            'n_calls': n_calls,
            'n_steps': n_steps,
            'n_output': y_out.shape[0],
            'integration': True,
            'fix_step': True,
            'dh': float(dh_abs.item()),
        }
        err_info = {
            'err_history': err_history,
            'max_step_error': 0.0,
        }
        return t_out, y_out, stats, err_info



import numpy as np
import torch
from . import boxlib
from .initial import Lattice

class spin:
    def __init__(self, theta, phi, lattice_type: Lattice,type="f32", device="gpu",thread:int=4):
        torch.set_num_threads(thread)
        data_type=boxlib.get_data_type(type)
        self.dtype=data_type
        self.device=boxlib.get_device(device)
        if isinstance(theta,torch.Tensor):
            self.theta=theta.to(device=self.device,dtype=self.dtype)
        else:
            self.theta = torch.tensor(theta,dtype=self.dtype,device=self.device)

        if isinstance(phi,torch.Tensor):
            self.phi=phi.to(device=self.device,dtype=self.dtype)
        else:
            self.phi = torch.tensor(phi,dtype=self.dtype,device=self.device)

        self.theta=self.theta.view(-1,1)
        self.phi=self.phi.view(-1,1)
        l_theta=len(self.theta)
        l_phi=len(self.phi)
        if l_theta!=l_phi:
            raise ValueError("theta and phi must have the same length")
        self.num = l_theta

        self.ang=torch.cat([self.theta,self.phi],1)
        self.ang=self.ang.reshape(-1,1)
        self.c_theta=torch.cos(self.theta)
        self.s_theta=torch.sin(self.theta)
        self.c_phi=torch.cos(self.phi)
        self.s_phi=torch.sin(self.phi)

        self.x, self.y, self.z = self.cart()
        self.cart_S=torch.cat([self.x,self.y,self.z],1)
        self.lattice_type=lattice_type

        num=lattice_type.size
        if self.num != np.prod(num):
            raise ValueError(f"initial condition theta and phi length {self.num} does not match lattice size {num}")
    def cart(self): 
        # convert to cartesian coordinates
        x = self.s_theta * self.c_phi
        y = self.s_theta * self.s_phi
        z = self.c_theta
        return x, y, z
    def update(self):
        self.c_theta=torch.cos(self.theta)
        self.s_theta=torch.sin(self.theta)
        self.c_phi=torch.cos(self.phi)
        self.s_phi=torch.sin(self.phi)
        
        self.x, self.y, self.z = self.cart()
        self.cart_S=torch.cat([self.x,self.y,self.z],1)




from magbox.solver.sde_solver import sde_solver
from .vec_normalize import vec_normalize
import torch

class sde3_solver(sde_solver):
    def _after_process(self, y: torch.Tensor) -> torch.Tensor:
        return vec_normalize(y)


import torch
from magbox.boxlib import get_Jmtx
from magbox import Lattice
import matplotlib.pyplot as plt

# warnings.filterwarnings("ignore", category=UserWarning, message=".*iCCP.*")

lt = Lattice(type="square", size=[8, 6, 4], periodic=True, J_direction=None)
tmp=get_Jmtx(lt)
print(tmp)
plt.figure(figsize=(6,6))
plt.spy(tmp.to_dense().cpu())
plt.show()

# def test_get_Jmtx():
#     ...
# def test_ode_rk45(


import numpy as np 
from thermal_tools import run_thermal, plot_thermal
from pathlib import Path

# Temp_all=np.logspace(np.log10(0.1),np.log10(10),4)
# dt_all=[0.01,0.1,0.5,1.0]
# alpha_all=[0.01,0.1,0.2,0.3]

# gamma_all=[0.5,1.0,2.0]

# for i,Temp in enumerate(Temp_all):
#     for j,dt in enumerate(dt_all):
#         for k,alpha in enumerate(alpha_all):
#             for l,gamma in enumerate(gamma_all):
#                 id=i*1000+j*100+k*10+l+1
#                 print(f"Running test id={id} with Temp={Temp}, dt={dt}, alpha={alpha}, gamma={gamma}")
                
#                 file=run_thermal(dt=dt, alpha=alpha, Temp=Temp,
#                             gamma=gamma, T=400, run_id=id)
#                 plot_thermal(file, start_time=40, bins=80)
#                 print(f"Completed test id={id} with Temp={Temp}, dt={dt}, alpha={alpha}, gamma={gamma}\n")

file=run_thermal(dt=0.1, alpha=0.01, Temp=5e-2, T=200, dtype='f64', device='gpu', spin_num=2**10, rtol=5e-2, atol=1e-6, fix_step=True, dh=None)

file=Path("./thermal_test_1.npz")
plot_thermal(file, start_time=50, bins=400,fit_counts=250)


import importlib.metadata
try:
	__version__ = importlib.metadata.version(__name__)
except importlib.metadata.PackageNotFoundError:
	# When running from a source checkout (not installed as a distribution).
	__version__ = "0.0.0"
__author__="Yutian Wang"

from .llg import llg
from .heff import heff
from .spin import spin

from .llg3 import llg3
from .heff3 import heff3
from .spin3 import spin3
from .boxlib import get_data_type
from .initial import Lattice, Vars

__all__ = ["llg", "heff", "spin", "llg3", "heff3", "spin3", "get_data_type", "Lattice", "Vars"]


from .heff3 import heff3
import torch 
from . import boxlib
from .initial import Lattice, Vars
from .spin3 import spin3
import warnings
from typing import Union, Tuple

class llg3:
    def __init__(self,x,y,z, lattice_type:Lattice,vars:Vars=Vars(),
                 dtype="f32", device="gpu", thread:int=4, require_ini_grad:bool=False,
                 gamma=1., alpha=0.01, Temp=0., dt=0.1, T=50., rtol:Union[float,None]=None, atol:Union[float,None]=None,
                 fix_step: bool = False, dh=None):
        warnings.filterwarnings("ignore", message="Sparse CSR tensor support is in beta state")
        sp = spin3(x,y,z,lattice_type,dtype=dtype, device=device, thread=thread, require_ini_grad=require_ini_grad)
        self.spin = sp

        self.rtol=rtol
        self.atol=atol
        self.fix_step=fix_step
        self.dh=dh
        dtype=sp.dtype
        device=sp.device
        self.num=sp.num

        self.gamma=torch.as_tensor(gamma,dtype=dtype,device=device)
        self.alpha=torch.as_tensor(alpha,dtype=dtype,device=device)
        self.Temp=torch.as_tensor(Temp,dtype=dtype,device=device)
        self.dt=torch.as_tensor(dt,dtype=dtype,device=device)
        self.T=torch.as_tensor(T,dtype=dtype,device=device)
        self.tspan=torch.linspace(0,self.T,int(self.T/self.dt)+1,dtype=dtype,device=device)
        self.h_fun=heff3(sp,vars)
        self.dtype=sp.dtype
        self.device=sp.device
        self.eps=torch.finfo(self.dtype).eps

        self.prefactor=-self.gamma/(1+self.alpha**2)

        # strength for thermal field
        self.thermal_strength=torch.sqrt(2*self.alpha*self.Temp/self.gamma)

    def llg_drift(self,t, S):
        x,y,z=spin3.get_xyz(S.view(-1,1))
        h3=self.h_fun.all3(t,x,y,z).view(-1,3)
        M1=torch.linalg.cross(S,h3,dim=1)
        M2=torch.linalg.cross(S,M1,dim=1)
        drift_core=M1 + self.alpha * M2
        return self.prefactor*drift_core
    
    def llg_thermal_no_correction(self, t, S):
        x,y,z=spin3.get_xyz(S.view(-1,1))
        h3=self.h_fun.all3(t,x,y,z).view(-1,3) 
        dim=h3.shape
        M1=torch.linalg.cross(S,h3,dim=1)
        M2=torch.linalg.cross(S,M1,dim=1)
        f=self.prefactor * (M1 + self.alpha * M2)
        def gfun(r):
            gM1=torch.linalg.cross(S,r,dim=1)
            gM2=torch.linalg.cross(S,gM1,dim=1)
            return self.thermal_strength*self.prefactor * (gM1 + self.alpha * gM2)
        return f, gfun, dim
    def llg_thermal(self, t, S):
        f, g, dim =self.llg_thermal_no_correction(t,S)
        correction = self.Stratonovich_correction(S) 
        return f-correction, g, dim
    def Stratonovich_correction(self,S):
        return 2*self.prefactor*self.alpha*self.Temp*S
    def run(self, ini=None, waitbar: bool = True) -> Tuple[torch.Tensor, torch.Tensor, dict, dict]:
        # error control
        if ini is None:
            ini=self.spin.cart_S
        if self.rtol is None:
            if self.Temp ==0 :
                rtol=max(self.alpha.item()*1e-2,1e-3)
            else:
                rtol=max(self.alpha.item()/5,1e-2)
        else:
            rtol=self.rtol

        if self.atol is None:
            if self.Temp ==0 :
                atol=max(self.alpha.item()*1e-4,1e-6)
            else:
                atol=max(self.alpha.item()*1e-2,1e-3)
        else:
            atol=self.atol
        odeset = {"rel_tol": rtol, "abs_tol": atol, "waitbar": waitbar, "fix_step": self.fix_step}
        if self.dh is not None:
            odeset["dh"] = self.dh
        
        if self.Temp==0:
            llg_fun=self.llg_drift
            t,Sout,stats,erro_info=boxlib.ode3_rk45(llg_fun, self.tspan, ini, options=odeset)
        else:
            llg_fun=self.llg_thermal_no_correction
            t,Sout,stats,erro_info=boxlib.ode3_sde_em(llg_fun, self.tspan,ini, options=odeset)
        return t,Sout,stats,erro_info



import torch
import warnings
import matplotlib.pyplot as plt
import numpy as np
from magbox.boxlib import ode_rk45
# from magbox.boxlib import ode_rk45 
def sin_ode(t, x):
    """
    定义微分方程: dx/dt = sin(x)
    """
    # return torch.sin(x)
    # return torch.sin(t)
    # return x**2
    return x
def out_fun(t,x0):
    
    # return 1/2*(t**2)+x0
    # return 1+x0-np.cos(t)
    # return -x0/(t*x0-1)
    return np.exp(t)*x0
   

def test_sin_ode():
    """
    测试求解 dx/dt = sin(x)
    """
    print("测试求解微分方程: dx/dt = sin(x)")
    
    # 设置初始条件和时间范围
    # tspan = torch.tensor([0,10])
    tspan=torch.linspace(0,50,1001,dtype=torch.float64)
    # x0 = torch.tensor([0.05])  # 初始条件 x(0) 
    x0=torch.tensor([1e15],dtype=torch.float64)
    
    # 设置求解器选项
    fxoptions = {
        'rel_tol': 1e-4,
        'abs_tol': 1e-6,
        'MaxStep': 5,
        'waitbar': True
    }
    
    print(f"初始条件: x(0) = {x0.item()}")
    print(f"时间范围: {tspan[0].item()} 到 {tspan[-1].item()}")
    
    # 求解微分方程
    t, x, stats, errInfo = ode_rk45(sin_ode, tspan, x0, fxoptions)
    
    print(f"\n求解完成!")
    print(f"最终解: x({t[-1]:.2f}) = {x[0, -1]:.6f}")
    print(f"函数调用次数: {stats['n_calls']}")
    print(f"求解段: {stats['n_steps']}")
    print(f"最大步长误差: {errInfo['max_step_error']:.2e}")
    
    return t, x, stats, errInfo

def plot_results(t, x,errInfo):
    """
    绘制结果
    """
    # 转换为numpy用于绘图
    t_np = t.detach().numpy()
    x_np = x.detach().numpy()

    x_theo = out_fun(t_np, x_np[:,0])
    
    # 创建图形
    plt.figure(figsize=(12, 8))
    
    # 绘制数值解
    plt.subplot(2, 2, 1)
    plt.plot(t_np, x_np[0], 'b-', linewidth=2, label='numeric')
    plt.plot(t_np,x_theo,'r--',label="theory")
    plt.xlabel('t')
    plt.ylabel('x(t)')
    plt.title('solution')
    plt.grid(True, alpha=0.3)
    plt.legend()
    
    # 绘制差距图 (x vs dx/dt)
    plt.subplot(2, 2, 2)
    plt.plot(t_np, np.abs(x_np/x_theo-1)[0], 'r-', linewidth=1)
    plt.xlabel('t')
    plt.ylabel('diff')
    plt.title('numeric-theory')
    plt.grid(True, alpha=0.3)
    
    # 绘制向量场
    plt.subplot(2, 2, 3)
    x_range = np.linspace(-1, 8, 20)
    t_range = np.linspace(0, 10, 10)
    T, X = np.meshgrid(t_range, x_range)
    
    # 计算向量场
    dX = np.sin(X)
    dT = np.ones_like(T)
    
    # 归一化箭头长度
    magnitude = np.sqrt(dT**2 + dX**2)
    dT_norm = dT / magnitude
    dX_norm = dX / magnitude
    
    plt.quiver(T, X, dT_norm, dX_norm, color='gray', alpha=0.6)
    plt.plot(t_np, x_np[0], 'b-', linewidth=2, label='solutio orbit')
    plt.xlabel('t')
    plt.ylabel('x')
    plt.title('vector field and solution orbit')
    plt.grid(True, alpha=0.3)
    plt.legend()
    
    # 绘制误差分析（如果有）
    plt.subplot(2, 2, 4)
    # 计算数值导数
    
    error = errInfo.get('err_history')
    plt.semilogy(np.linspace(1,len(error),len(error)), error, 'g-', alpha=0.7)
    plt.xlabel('step')
    plt.ylabel('erro')
    plt.title('erro')
    plt.grid(True, alpha=0.3)
    
    plt.tight_layout()
    plt.show()

def compare_different_initial_conditions():
    """
    比较不同初始条件的解
    """
    print("\n比较不同初始条件的解...")
    
    tspan = torch.tensor([0.0, 10.0])
    initial_conditions = [0.05, 0.11, 2.0, 3.0]
    
    fxoptions = {
        'rel_tol': 1e-3, 
        'abs_tol': 1e-6,
        'waitbar': True
    }
    
    plt.figure(figsize=(10, 6))
    
    for i, x0_val in enumerate(initial_conditions):
        x0 = torch.tensor([x0_val])
        t, x, stats, errInfo = ode_rk45(sin_ode, tspan, x0, fxoptions)
        
        t_np = t.detach().numpy()
        x_np = x.detach().numpy()
        
        plt.plot(t_np, x_np[0], linewidth=2, 
                label=f'x(0) = {x0_val}')
        
        print(f"初始条件 x(0) = {x0_val:.1f}: 最终值 x(10) = {x_np[0, -1]:.4f}")
    
    plt.xlabel('t')
    plt.ylabel('x(t)')
    plt.title('solution with different init')
    plt.grid(True, alpha=0.3)
    plt.legend()
    plt.tight_layout()
    plt.show()
if __name__ == "__main__":
    # 运行测试
    warnings.filterwarnings("ignore", message="libpng")
    print("=" * 50)
    print("ode_rk45 测试程序: dx/dt = sin(x)")
    print("=" * 50)
    
    # 基本测试
    t, x, stats, errInfo = test_sin_ode()
    
    # 绘制结果
    plot_results(t, x, errInfo)
    
    # 比较不同初始条件
    compare_different_initial_conditions()
    
    print("\n所有测试完成!")